/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil -*- */
/*
 * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2010 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * Copyright (c) 2007-2010 Oracle and/or its affiliates.  All rights reserved.
 * Copyright (c) 2015      Research Organization for Information Science
 *                         and Technology (RIST). All rights reserved.
 * Copyright (c) 2015      Los Alamos National Security, LLC. All rights
 *                         reserved.
 * Copyright (c) 2017      Intel, Inc. All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#ifndef PMIX_SYS_ARCH_ATOMIC_H
#define PMIX_SYS_ARCH_ATOMIC_H 1

/*
 * On ia32, we use cmpxchg.
 */

#define PMIXSMPLOCK "lock; "
#define PMIXMB() __asm__ __volatile__("": : :"memory")


/**********************************************************************
 *
 * Define constants for IA32
 *
 *********************************************************************/
#define PMIX_HAVE_ATOMIC_MEM_BARRIER 1

#define PMIX_HAVE_ATOMIC_CMPSET_32 1

#define PMIX_HAVE_ATOMIC_MATH_32 1
#define PMIX_HAVE_ATOMIC_ADD_32 1
#define PMIX_HAVE_ATOMIC_SUB_32 1

#define PMIX_HAVE_ATOMIC_CMPSET_64 1

#undef PMIX_HAVE_INLINE_ATOMIC_CMPSET_64
#define PMIX_HAVE_INLINE_ATOMIC_CMPSET_64 0

/**********************************************************************
 *
 * Memory Barriers
 *
 *********************************************************************/
#if PMIX_GCC_INLINE_ASSEMBLY

static inline void pmix_atomic_mb(void)
{
    PMIXMB();
}


static inline void pmix_atomic_rmb(void)
{
    PMIXMB();
}


static inline void pmix_atomic_wmb(void)
{
    PMIXMB();
}

static inline void pmix_atomic_isync(void)
{
}

#endif /* PMIX_GCC_INLINE_ASSEMBLY */


/**********************************************************************
 *
 * Atomic math operations
 *
 *********************************************************************/
#if PMIX_GCC_INLINE_ASSEMBLY

static inline int pmix_atomic_cmpset_32(volatile int32_t *addr,
                                        int32_t oldval,
                                        int32_t newval)
{
   unsigned char ret;
   __asm__ __volatile__ (
                       PMIXSMPLOCK "cmpxchgl %3,%2   \n\t"
                               "sete     %0      \n\t"
                       : "=qm" (ret), "+a" (oldval), "+m" (*addr)
                       : "q"(newval)
                       : "memory", "cc");

   return (int)ret;
}

#endif /* PMIX_GCC_INLINE_ASSEMBLY */

#define pmix_atomic_cmpset_acq_32 pmix_atomic_cmpset_32
#define pmix_atomic_cmpset_rel_32 pmix_atomic_cmpset_32

#if PMIX_GCC_INLINE_ASSEMBLY

#if 0

/* some versions of GCC won't let you use ebx period (even though they
   should be able to save / restore for the life of the inline
   assembly).  For the beta, just use the non-inline version */

#ifndef ll_low /* GLIBC provides these somewhere, so protect */
#define ll_low(x)       *(((unsigned int*)&(x))+0)
#define ll_high(x)      *(((unsigned int*)&(x))+1)
#endif

/* On Linux the EBX register is used by the shared libraries
 * to keep the global offset. In same time this register is
 * required by the cmpxchg8b instruction (as an input parameter).
 * This conflict force us to save the EBX before the cmpxchg8b
 * and to restore it afterward.
 */
static inline int pmix_atomic_cmpset_64(volatile int64_t *addr,
                                        int64_t oldval,
                                        int64_t newval)
{
   /*
    * Compare EDX:EAX with m64. If equal, set ZF and load ECX:EBX into
    * m64. Else, clear ZF and load m64 into EDX:EAX.
    */
    unsigned char ret;

    __asm__ __volatile__(
                    "push %%ebx            \n\t"
                    "movl %4, %%ebx        \n\t"
                    SMPLOCK "cmpxchg8b (%1)  \n\t"
                    "sete %0               \n\t"
                    "pop %%ebx             \n\t"
                    : "=qm"(ret)
                    : "D"(addr), "a"(ll_low(oldval)), "d"(ll_high(oldval)),
                      "r"(ll_low(newval)), "c"(ll_high(newval))
                    : "cc", "memory", "ebx");
    return (int) ret;
}
#endif /* if 0 */

#endif /* PMIX_GCC_INLINE_ASSEMBLY */

#define pmix_atomic_cmpset_acq_64 pmix_atomic_cmpset_64
#define pmix_atomic_cmpset_rel_64 pmix_atomic_cmpset_64

#if PMIX_GCC_INLINE_ASSEMBLY

#define PMIX_HAVE_ATOMIC_SWAP_32 1

static inline int32_t pmix_atomic_swap_32( volatile int32_t *addr,
                                           int32_t newval)
{
    int32_t oldval;

    __asm__ __volatile__("xchg %1, %0" :
                         "=r" (oldval), "=m" (*addr) :
                         "0" (newval), "m" (*addr) :
                         "memory");
    return oldval;
}

#endif /* PMIX_GCC_INLINE_ASSEMBLY */


#if PMIX_GCC_INLINE_ASSEMBLY

/**
 * atomic_add - add integer to atomic variable
 * @i: integer value to add
 * @v: pointer of type int
 *
 * Atomically adds @i to @v.
 */
static inline int32_t pmix_atomic_add_32(volatile int32_t* v, int i)
{
    int ret = i;
   __asm__ __volatile__(
                        PMIXSMPLOCK "xaddl %1,%0"
                        :"+m" (*v), "+r" (ret)
                        :
                        :"memory", "cc"
                        );
   return (ret+i);
}


/**
 * atomic_sub - subtract the atomic variable
 * @i: integer value to subtract
 * @v: pointer of type int
 *
 * Atomically subtracts @i from @v.
 */
static inline int32_t pmix_atomic_sub_32(volatile int32_t* v, int i)
{
    int ret = -i;
   __asm__ __volatile__(
                        PMIXSMPLOCK "xaddl %1,%0"
                        :"+m" (*v), "+r" (ret)
                        :
                        :"memory", "cc"
                        );
   return (ret-i);
}

#endif /* PMIX_GCC_INLINE_ASSEMBLY */

#endif /* ! PMIX_SYS_ARCH_ATOMIC_H */
